@@ -1,10 +1,12 @@
 # flake8: noqa
+from .class_validators import validator
 from .env_settings import BaseSettings
 from .error_wrappers import ValidationError
 from .errors import *
-from .fields import Required, Schema
-from .main import BaseConfig, BaseModel, create_model, validate_model, validator
+from .fields import Required
+from .main import BaseConfig, BaseModel, create_model, validate_model
 from .parse import Protocol
 from .types import *
 from .version import VERSION
 from . import dataclasses
+from .schema import Schema
