@@ -1,23 +1,17 @@
-import inspect
-from enum import Enum, IntEnum
-from typing import Any, Callable, List, Mapping, NamedTuple, Pattern, Set, Tuple, Type, Union
+import warnings
+from enum import IntEnum
+from typing import Any, Dict, List, Mapping, Optional, Pattern, Set, Tuple, Type, Union
 
 from . import errors as errors_
+from .class_validators import Validator, ValidatorSignature, get_validator_signature
 from .error_wrappers import ErrorWrapper
 from .types import Json, JsonWrapper
-from .utils import display_as_type, list_like
-from .validators import NoneType, dict_validator, find_validators, not_none_validator
+from .utils import ForwardRef, display_as_type, lenient_issubclass, list_like
+from .validators import NoneType, dict_validator, find_validators, is_none_validator
 
 Required: Any = Ellipsis
 
 
-class ValidatorSignature(IntEnum):
-    JUST_VALUE = 1
-    VALUE_KWARGS = 2
-    CLS_JUST_VALUE = 3
-    CLS_VALUE_KWARGS = 4
-
-
 class Shape(IntEnum):
     SINGLETON = 1
     LIST = 2
@@ -26,29 +20,6 @@ class Shape(IntEnum):
     TUPLE = 5
 
 
-class Validator(NamedTuple):
-    func: Callable
-    pre: bool
-    whole: bool
-    always: bool
-    check_fields: bool
-
-
-class Schema:
-    """
-    Used to provide extra information about a field in a model schema.
-    """
-
-    __slots__ = 'default', 'alias', 'title', 'choice_names', 'extra'
-
-    def __init__(self, default, *, alias=None, title=None, choice_names=None, **extra):
-        self.default = default
-        self.alias = alias
-        self.title = title
-        self.choice_names = choice_names
-        self.extra = extra
-
-
 class Field:
     __slots__ = (
         'type_',
@@ -63,7 +34,7 @@ class Field:
         'name',
         'alias',
         'has_alias',
-        '_schema',
+        'schema',
         'validate_always',
         'allow_none',
         'shape',
@@ -76,38 +47,40 @@ class Field:
         *,
         name: str,
         type_: Type,
-        class_validators: List[Validator],
-        default: Any,
-        required: bool,
+        class_validators: Optional[Dict[str, Validator]],
         model_config: Any,
+        default: Any = None,
+        required: bool = True,
         alias: str = None,
-        allow_none: bool = False,
-        schema: Schema = None,
+        schema=None,
     ):
 
         self.name: str = name
         self.has_alias: bool = bool(alias)
         self.alias: str = alias or name
         self.type_: type = type_
-        self.class_validators = class_validators or []
+        self.class_validators = class_validators or {}
+        self.default: Any = default
+        self.required: bool = required
+        self.model_config = model_config
+        self.schema: 'schema.Schema' = schema
+
+        self.allow_none: bool = False
         self.validate_always: bool = False
         self.sub_fields: List[Field] = None
         self.key_field: Field = None
         self.validators = []
         self.whole_pre_validators = None
         self.whole_post_validators = None
-        self.default: Any = default
-        self.required: bool = required
-        self.model_config = model_config
-        self.allow_none: bool = allow_none
         self.parse_json: bool = False
         self.shape: Shape = Shape.SINGLETON
-        self._schema: Schema = schema
         self.prepare()
 
     @classmethod
     def infer(cls, *, name, value, annotation, class_validators, config):
         schema_from_config = config.get_field_schema(name)
+        from .schema import Schema, get_annotation_from_schema
+
         if isinstance(value, Schema):
             schema = value
             value = schema.default
@@ -115,6 +88,7 @@ class Field:
             schema = Schema(value, **schema_from_config)
         schema.alias = schema.alias or schema_from_config.get('alias')
         required = value == Required
+        annotation = get_annotation_from_schema(annotation, schema)
         return cls(
             name=name,
             type_=annotation,
@@ -130,8 +104,8 @@ class Field:
         self.model_config = config
         schema_from_config = config.get_field_schema(self.name)
         if schema_from_config:
-            self._schema.alias = self._schema.alias or schema_from_config.get('alias')
-            self.alias = self._schema.alias
+            self.schema.alias = self.schema.alias or schema_from_config.get('alias')
+            self.alias = self.schema.alias
 
     @property
     def alt_alias(self):
@@ -144,70 +118,24 @@ class Field:
         if self.type_ is None:
             raise errors_.ConfigError(f'unable to infer type for attribute "{self.name}"')
 
+        if type(self.type_) == ForwardRef:
+            # self.type_ is currently a ForwardRef and there's nothing we can do now,
+            # user will need to call model.update_forward_refs()
+            return
+
         self.validate_always: bool = (
-            getattr(self.type_, 'validate_always', False) or any(v.always for v in self.class_validators)
+            getattr(self.type_, 'validate_always', False) or any(v.always for v in self.class_validators.values())
         )
 
-        if not self.required and not self.validate_always and self.default is None:
+        if not self.required and self.default is None:
             self.allow_none = True
 
         self._populate_sub_fields()
         self._populate_validators()
 
-    def schema(self, by_alias=True):
-        s = dict(title=self._schema.title or self.alias.title(), required=self.required)
-
-        if not self.required and self.default is not None:
-            s['default'] = self.default
-        s.update(self._schema.extra)
-
-        ts = self.type_schema(by_alias)
-        s.update(ts if isinstance(ts, dict) else {'type': ts})
-        return s
-
-    def type_schema(self, by_alias):
-        if self.shape is Shape.LIST:
-            return {'type': 'list', 'item_type': self._singleton_schema(by_alias)}
-        if self.shape is Shape.SET:
-            return {'type': 'set', 'item_type': self._singleton_schema(by_alias)}
-        elif self.shape is Shape.MAPPING:
-            return {
-                'type': 'mapping',
-                'item_type': self._singleton_schema(by_alias),
-                'key_type': self.key_field.type_schema(by_alias),
-            }
-        elif self.shape is Shape.TUPLE:
-            return {'type': 'tuple', 'item_types': [sf.type_schema(by_alias) for sf in self.sub_fields]}
-        else:
-            assert self.shape is Shape.SINGLETON, self.shape
-            return self._singleton_schema(by_alias)
-
-    def _singleton_schema(self, by_alias):
-        if self.sub_fields:
-            if len(self.sub_fields) == 1:
-                return self.sub_fields[0].type_schema(by_alias)
-            else:
-                return {'type': 'any_of', 'types': [sf.type_schema(by_alias) for sf in self.sub_fields]}
-        elif self.type_ is Any:
-            return 'any'
-        elif issubclass(self.type_, Enum):
-            choice_names = self._schema.choice_names or {}
-            return {
-                'type': display_as_type(self.type_),
-                'choices': [
-                    (v.value, choice_names.get(v.value) or k.title()) for k, v in self.type_.__members__.items()
-                ],
-            }
-
-        type_schema_method = getattr(self.type_, 'type_schema', None)
-        if callable(type_schema_method):
-            return type_schema_method(by_alias)
-        else:
-            return display_as_type(self.type_)
-
-    def _populate_sub_fields(self):  # noqa: C901 (ignore complexity)
+    def _populate_sub_fields(self):
         # typing interface is horrible, we have to do some ugly checks
-        if isinstance(self.type_, type) and issubclass(self.type_, JsonWrapper):
+        if lenient_issubclass(self.type_, JsonWrapper):
             self.type_ = self.type_.inner_type
             self.parse_json = True
 
@@ -224,8 +152,7 @@ class Field:
                 if type_ is NoneType:
                     self.allow_none = True
                     self.required = False
-                else:
-                    types_.append(type_)
+                types_.append(type_)
             self.sub_fields = [self._create_sub_type(t, f'{self.name}_{display_as_type(t)}') for t in types_]
             return
 
@@ -242,7 +169,7 @@ class Field:
             self.shape = Shape.SET
         else:
             assert issubclass(origin, Mapping)
-            self.key_field = self._create_sub_type(self.type_.__args__[0], 'key_' + self.name)
+            self.key_field = self._create_sub_type(self.type_.__args__[0], 'key_' + self.name, for_keys=True)
             self.type_ = self.type_.__args__[1]
             self.shape = Shape.MAPPING
 
@@ -250,50 +177,50 @@ class Field:
             # type_ has been refined eg. as the type of a List and sub_fields needs to be populated
             self.sub_fields = [self._create_sub_type(self.type_, '_' + self.name)]
 
-    def _create_sub_type(self, type_, name):
+    def _create_sub_type(self, type_, name, *, for_keys=False):
         return self.__class__(
             type_=type_,
             name=name,
-            class_validators=self.class_validators,
-            default=self.default,
-            required=self.required,
-            allow_none=self.allow_none,
+            class_validators=None if for_keys else {k: v for k, v in self.class_validators.items() if not v.whole},
             model_config=self.model_config,
         )
 
     def _populate_validators(self):
+        class_validators_ = self.class_validators.values()
         if not self.sub_fields:
-            get_validators = getattr(self.type_, 'get_validators', None)
+            get_validators = getattr(self.type_, '__get_validators__', None)
+            if not get_validators:
+                get_validators = getattr(self.type_, 'get_validators', None)
+                if get_validators:
+                    warnings.warn(
+                        f'get_validators has been replaced by __get_validators__ (on {self.name})', DeprecationWarning
+                    )
             v_funcs = (
-                *tuple(v.func for v in self.class_validators if not v.whole and v.pre),
+                *tuple(v.func for v in class_validators_ if not v.whole and v.pre),
                 *(
                     get_validators()
                     if get_validators
                     else find_validators(self.type_, self.model_config.arbitrary_types_allowed)
                 ),
-                *tuple(v.func for v in self.class_validators if not v.whole and not v.pre),
+                *tuple(v.func for v in class_validators_ if not v.whole and not v.pre),
             )
             self.validators = self._prep_vals(v_funcs)
 
-        if self.class_validators:
-            self.whole_pre_validators = self._prep_vals(v.func for v in self.class_validators if v.whole and v.pre)
-            self.whole_post_validators = self._prep_vals(v.func for v in self.class_validators if v.whole and not v.pre)
+        if class_validators_:
+            self.whole_pre_validators = self._prep_vals(v.func for v in class_validators_ if v.whole and v.pre)
+            self.whole_post_validators = self._prep_vals(v.func for v in class_validators_ if v.whole and not v.pre)
 
-    def _prep_vals(self, v_funcs):
-        v = []
-        for f in v_funcs:
-            if not f or (self.allow_none and f is not_none_validator):
-                continue
-            v.append((_get_validator_signature(f), f))
-        return tuple(v)
+    @staticmethod
+    def _prep_vals(v_funcs):
+        return tuple((get_validator_signature(f), f) for f in v_funcs if f)
 
-    def validate(self, v, values, *, loc, cls=None):  # noqa: C901 (ignore complexity)
-        if self.allow_none and v is None:
+    def validate(self, v, values, *, loc, cls=None):
+        if self.allow_none and not self.validate_always and v is None:
             return None, None
 
         loc = loc if isinstance(loc, tuple) else (loc,)
 
-        if self.parse_json:
+        if v is not None and self.parse_json:
             v, error = self._validate_json(v, loc)
             if error:
                 return v, error
@@ -425,6 +352,24 @@ class Field:
                 return v, ErrorWrapper(exc, loc=loc, config=self.model_config)
         return v, None
 
+    def include_in_schema(self) -> bool:
+        """
+        False if this is a simple field just allowing None as used in Unions/Optional.
+        """
+        return len(self.validators) > 1 or self.validators[0][1] != is_none_validator
+
+    def is_complex(self):
+        """
+        Whether the field is "complex" eg. env variables should be parsed as JSON.
+        """
+        from .main import BaseModel
+
+        return (
+            self.shape != Shape.SINGLETON
+            or lenient_issubclass(self.type_, (BaseModel, list, set, dict))
+            or hasattr(self.type_, '__pydantic_model__')  # pydantic dataclass
+        )
+
     def __repr__(self):
         return f'<Field({self})>'
 
@@ -439,32 +384,3 @@ class Field:
         if self.alt_alias:
             parts.append('alias=' + self.alias)
         return ' '.join(parts)
-
-
-def _get_validator_signature(validator):
-    signature = inspect.signature(validator)
-
-    # bind here will raise a TypeError so:
-    # 1. we can deal with it before validation begins
-    # 2. (more importantly) it doesn't get confused with a TypeError when executing the validator
-    try:
-        if 'cls' in signature._parameters:
-            if len(signature.parameters) == 2:
-                signature.bind(object(), 1)
-                return ValidatorSignature.CLS_JUST_VALUE
-            else:
-                signature.bind(object(), 1, values=2, config=3, field=4)
-                return ValidatorSignature.CLS_VALUE_KWARGS
-        else:
-            if len(signature.parameters) == 1:
-                signature.bind(1)
-                return ValidatorSignature.JUST_VALUE
-            else:
-                signature.bind(1, values=2, config=3, field=4)
-                return ValidatorSignature.VALUE_KWARGS
-    except TypeError as e:
-        raise errors_.ConfigError(
-            f'Invalid signature for validator {validator}: {signature}, should be: '
-            f'(value) or (value, *, values, config, field) or for class validators '
-            f'(cls, value) or (cls, value, *, values, config, field)'
-        ) from e
