@@ -1,11 +1,12 @@
 import inspect
 import re
+import sys
 from contextlib import contextmanager
 from enum import Enum
 from functools import lru_cache
 from importlib import import_module
 from textwrap import dedent
-from typing import List, Pattern, Tuple, Type
+from typing import List, Pattern, Tuple, Type, _eval_type
 
 from . import errors
 
@@ -19,6 +20,11 @@ try:
 except ImportError:
     from typing import _Final as typing_base
 
+try:
+    from typing import ForwardRef
+except ImportError:
+    # python 3.6
+    ForwardRef = None
 
 PRETTY_REGEX = re.compile(r'([\w ]*?) *<(.*)> *')
 
@@ -129,7 +135,7 @@ def display_as_type(v):
     if not isinstance(v, typing_base) and not isinstance(v, type):
         v = type(v)
 
-    if isinstance(v, type) and issubclass(v, Enum):
+    if lenient_issubclass(v, Enum):
         if issubclass(v, int):
             return 'int'
         elif issubclass(v, str):
@@ -201,3 +207,42 @@ def url_regex_generator(*, relative: bool, require_tld: bool) -> Pattern:
         ),
         re.IGNORECASE,
     )
+
+
+def lenient_issubclass(cls, class_or_tuple):
+    return isinstance(cls, type) and issubclass(cls, class_or_tuple)
+
+
+def in_ipython():
+    """
+    Check whether we're in an ipython environment, including jupyter notebooks.
+    """
+    try:
+        __IPYTHON__
+    except NameError:
+        return False
+    else:  # pragma: no cover
+        return True
+
+
+def resolve_annotations(raw_annotations, module):
+    """
+    Partially taken from typing.get_type_hints.
+
+    Resolve string or ForwardRef annotations into type objects if possible.
+    """
+    if module:
+        base_globals = sys.modules[module].__dict__
+    else:
+        base_globals = None
+    annotations = {}
+    for name, value in raw_annotations.items():
+        if isinstance(value, str):
+            value = ForwardRef(value, is_argument=False)
+        try:
+            value = _eval_type(value, base_globals, None)
+        except NameError:
+            # this is ok, it can be fixed with update_forward_refs
+            pass
+        annotations[name] = value
+    return annotations
